#!/usr/bin/env python

import os
import sqlite3
import binascii
import itertools
import logging
log = logging.getLogger('gglsbl3')


class StorageBase(object):
    @staticmethod
    def compress_ranges(nums):
        """Compress consequtive ranges in sequence of numbers

        E.g. [1,2,3,4,7] -> '1-4,7'
        """
        if not nums:
            return None
        nums.sort()
        buf = []
        buf.append(nums[0])
        for i in range(1, len(nums)):
            if nums[i-1] == nums[i]:
                pass
            elif nums[i] - nums[i-1] == 1:
                if buf[-1] is not None:
                    buf.append(None)
            else:
                if buf[-1] is None:
                    buf.append(nums[i-1])
                buf.append(nums[i])
        if buf[-1] is None:
            buf.append(nums[-1])
        output = ','.join([str(i) for i in buf])
        output = output.replace(',None,', '-')
        return output

    @staticmethod
    def expand_ranges(list_of_ranges):
        """Do the opposite of compress_ranges()

        E.g. ['1-4,7', '9-11'] -> [1,2,3,4,7,9,10,11]
        """
        nums = []
        for ranges in list_of_ranges:
            for chunk_range in ranges.strip().split(','):
                if type(chunk_range) is int:
                    nums.append(chunk_range)
                elif chunk_range.isdigit():
                    nums.append(int(chunk_range))
                else:
                    try:
                        r1, r2 = chunk_range.split('-')
                        r1 = int(r1)
                        r2 = int(r2) + 1
                        nums.extend(range(r1, r2))
                    except ValueError as e:
                        log.error('Failed to parse chunk range %s (%s)', chunk_range, e)
                        raise
        return nums


class SqliteStorage(StorageBase):
    """Storage abstraction for local GSB cache"""
    def __init__(self, db_path):
        do_init_db = not os.path.isfile(db_path)
        log.info('Opening SQLite DB %s' % db_path)
        self.db = sqlite3.connect(db_path)
        self.dbc = self.db.cursor()
        if do_init_db:
            log.info('SQLite DB does not exist, initializing')
            self.init_db()

    def init_db(self):
        self.dbc.execute(
        """CREATE TABLE chunk (
            chunk_number integer NOT NULL,
            timestamp timestamp DEFAULT current_timestamp,
            list_name character varying(127) NOT NULL,
            chunk_type TEXT CHECK( chunk_type IN ('add','sub')) NOT NULL,
            PRIMARY KEY (chunk_number, list_name, chunk_type)
            )"""
        )
        self.dbc.execute(
        """CREATE TABLE full_hash (
            value BLOB NOT NULL,
            list_name character varying(127) NOT NULL,
            metadata TINYINT,
            downloaded_at timestamp DEFAULT current_timestamp,
            expires_at timestamp without time zone NOT NULL,
            PRIMARY KEY (value)
            )"""
        )

        self.dbc.execute(
        """CREATE TABLE hash_prefix (
            value BLOB NOT NULL,
            chunk_number integer NOT NULL,
            timestamp timestamp without time zone DEFAULT current_timestamp,
            list_name character varying(127) NOT NULL,
            chunk_type TEXT CHECK( chunk_type IN ('add','sub')) NOT NULL,
            full_hash_expires_at timestamp NOT NULL DEFAULT current_timestamp,
            PRIMARY KEY (value, chunk_number, list_name, chunk_type),
            FOREIGN KEY(chunk_number, list_name, chunk_type)
                REFERENCES chunk(chunk_number, list_name, chunk_type)
                ON DELETE CASCADE
            )"""
        )
        self.db.commit()

    def chunk_exists(self, chunk):
        "Check if given chunk records already exist in the database"
        q = 'SELECT COUNT(*) FROM chunk WHERE chunk_number=? AND \
            chunk_type=? AND list_name=?'
        self.dbc.execute(q, [chunk.chunk_number, chunk.chunk_type, chunk.list_name])
        if self.dbc.fetchall()[0][0] > 0:
            return True
        return False

    def store_chunk(self, chunk):
        "Store chunk in the database"
        log.debug('Storing %s chunk #%s for list name %s' % (chunk.chunk_type, chunk.chunk_number, chunk.list_name))
        self.insert_chunk(chunk)
        for hash_value in chunk.hashes:
            hash_prefix = {
                'list_name': chunk.list_name,
                'chunk_number': chunk.chunk_number,
                'chunk_type': chunk.chunk_type,
                'value': sqlite3.Binary(hash_value),
            }
            self.insert_hash_prefix(hash_prefix)
        self.db.commit()

    def insert_chunk(self, chunk):
        "Insert hash prefixes from the chunk to the database"
        q = 'INSERT INTO chunk (chunk_number, list_name, chunk_type) \
            VALUES (?, ?, ?)'
        self.dbc.execute(q, [chunk.chunk_number, chunk.list_name, chunk.chunk_type])

    def insert_hash_prefix(self, hash_prefix):
        "Insert individual hash prefix to the database"
        q = 'INSERT INTO hash_prefix (value, chunk_number, list_name, chunk_type) \
            VALUES (?, ?, ?, ?)'
        params = [hash_prefix[k] for k in
                        ('value', 'chunk_number', 'list_name', 'chunk_type')]
        try:
            self.dbc.execute(q, params)
        except sqlite3.IntegrityError as e:
            log.warning("Trying to insert existing hash prefix: '%s' (%s)", hash_prefix, e)


    def store_full_hashes(self, hash_prefix, hashes):
        "Store hashes found for the given hash prefix"
        self.cleanup_expired_hashes()
        cache_lifetime = hashes['cache_lifetime']
        for hash_info, metadata_info in itertools.zip_longest(list(hashes['hashes'].items()), list(hashes['metadata'].items())):
            list_name_bytes = hash_info[0]
            #  log.debug("list name bytes is: {list_name_bytes}".format(list_name_bytes=list_name_bytes))
            list_name = list_name_bytes.decode("ascii")  # convert the bytestring listname to a 'normal' python string
            #  log.debug("list_name is {list_name}".format(list_name=list_name))
            hash_values = hash_info[1]
            metadata_values = metadata_info[1]
            log.debug("list_name: {list_name}".format(list_name=list_name))
            for hash_value, metadata_value in itertools.zip_longest(hash_values, metadata_values):
                if metadata_value is None:
                    metadata_value = 0
                log.debug("metadata value is: {metadata_value}".format(metadata_value=metadata_value))
                log.debug('storing hash: "{hash}"'.format(hash=binascii.hexlify(hash_value).decode("ascii")))
                q = "INSERT INTO full_hash (value, list_name, metadata, downloaded_at, expires_at)\
                    VALUES (?, ?, ?, current_timestamp, datetime(current_timestamp, '+%d SECONDS'))"
                q = q % cache_lifetime
                log.debug(q)
                self.dbc.execute(q, [sqlite3.Binary(hash_value), list_name, metadata_value])
        q = "UPDATE hash_prefix SET full_hash_expires_at=datetime(current_timestamp, '+%d SECONDS') \
            WHERE chunk_type='add' AND value=?"
        self.dbc.execute(q % cache_lifetime, [sqlite3.Binary(hash_prefix)])
        self.db.commit()

    def full_hash_sync_required(self, hash_prefix):
        """Check if hashes for the given hash prefix have expired

        and that prefix needs to be re-queried
        """
        q = "SELECT COUNT(*) FROM hash_prefix WHERE \
            full_hash_expires_at > current_timestamp AND chunk_type='add' AND value=?"
        self.dbc.execute(q, [sqlite3.Binary(hash_prefix)])
        c = self.dbc.fetchall()[0][0]
        return c == 0

    def lookup_full_hash(self, hash_value):
        "Query DB to see if hash is blacklisted"
        q = 'SELECT list_name, metadata FROM full_hash WHERE value=?'
        self.dbc.execute(q, [sqlite3.Binary(hash_value)])
        fetched = self.dbc.fetchall()
        log.debug("fetched from database: {fetched}".format(fetched=fetched))
        return [{"list": h[0], "metadata": h[1]} for h in fetched]

    def lookup_hash_prefix(self, hash_prefix):
        """Check if hash prefix is in the list and does not have 'sub'
        status signifying that it should be evicted from the blacklist
        """
        q = 'SELECT list_name FROM hash_prefix WHERE chunk_type=? AND value=?'
        self.dbc.execute(q, ['add', sqlite3.Binary(hash_prefix)])
        lists_add = [r[0] for r in self.dbc.fetchall()]
        if len(lists_add) == 0:
            return False
        self.dbc.execute(q, ['sub', sqlite3.Binary(hash_prefix)])
        lists_sub = [r[0] for r in self.dbc.fetchall()]
        if len(lists_sub) == 0:
            return True
        if set(lists_add) - set(lists_sub):
            return True
        return False

    def cleanup_expired_hashes(self):
        "Delete all hashes that behind their expiration date"
        q = 'DELETE FROM full_hash WHERE expires_at < current_timestamp'
        self.dbc.execute(q)
        self.db.commit()

    def del_add_chunks(self, chunk_numbers):
        "Delete records associated with 'add' chunk"
        if not chunk_numbers:
            return
        log.info('Deleting "add" chunks %s' % repr(chunk_numbers))
        for cn in self.expand_ranges(chunk_numbers):
            q = 'DELETE FROM chunk WHERE chunk_type=? AND chunk_number=?'
            self.dbc.execute(q, ['add', cn])
        self.db.commit()

    def del_sub_chunks(self, chunk_numbers):
        "Delete records associated with 'sub' chunk"
        if not chunk_numbers:
            return
        log.info('Deleting "sub" chunks %s' % repr(chunk_numbers))
        for cn in self.expand_ranges(chunk_numbers):
            q = 'DELETE FROM chunk WHERE chunk_type=? AND chunk_number=?'
            self.dbc.execute(q, ['sub', cn])
        self.db.commit()

    def get_existing_chunks(self):
        "Get the list of chunks that are available in the local cache"
        output = {}
        for chunk_type in ('add', 'sub'):
            q = "SELECT list_name, group_concat(chunk_number) FROM chunk \
                WHERE chunk_type=? GROUP BY list_name"
            self.dbc.execute(q, [chunk_type])
            for list_name, chunks in self.dbc.fetchall():
                if list_name not in output:
                    output[list_name] = {}
                chunks = [int(c) for c in chunks.split(',')]
                output[list_name][chunk_type] = self.compress_ranges(chunks)
        return output

    # FiXME: add stats about number of chunks, number of fullhashes

    def total_cleanup(self):
        "Reset local cache"
        q = 'DROP TABLE hash_prefix'
        self.dbc.execute(q)
        q = 'DROP TABLE chunk'
        self.dbc.execute(q)
        q = 'DROP TABLE full_hash'
        self.dbc.execute(q)
        self.db.commit()
        self.init_db()

    def close(self):
        self.db.close()



# Generated by the protocol buffer compiler.  DO NOT EDIT! - Oops. I did...
# source: MalwarePatternType.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
# from google.protobuf import symbol_database as _symbol_database  # had to comment this one because it doesnt exist in the protobuf py3 port
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

# _sym_db = _symbol_database.Default()  #  see why this is commented above




DESCRIPTOR = _descriptor.FileDescriptor(
  name='MalwarePatternType.proto',
  package='',
  serialized_pb=_b('\n\x18MalwarePatternType.proto\"{\n\x12MalwarePatternType\x12\x36\n\x0cpattern_type\x18\x01 \x02(\x0e\x32 .MalwarePatternType.PATTERN_TYPE\"-\n\x0cPATTERN_TYPE\x12\x0b\n\x07LANDING\x10\x01\x12\x10\n\x0c\x44ISTRIBUTION\x10\x02')
)
# _sym_db.RegisterFileDescriptor(DESCRIPTOR)



_MALWAREPATTERNTYPE_PATTERN_TYPE = _descriptor.EnumDescriptor(
  name='PATTERN_TYPE',
  full_name='MalwarePatternType.PATTERN_TYPE',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='LANDING', index=0, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DISTRIBUTION', index=1, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=106,
  serialized_end=151,
)
# _sym_db.RegisterEnumDescriptor(_MALWAREPATTERNTYPE_PATTERN_TYPE)


_MALWAREPATTERNTYPE = _descriptor.Descriptor(
  name='MalwarePatternType',
  full_name='MalwarePatternType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pattern_type', full_name='MalwarePatternType.pattern_type', index=0,
      number=1, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _MALWAREPATTERNTYPE_PATTERN_TYPE,
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=28,
  serialized_end=151,
)

_MALWAREPATTERNTYPE.fields_by_name['pattern_type'].enum_type = _MALWAREPATTERNTYPE_PATTERN_TYPE
_MALWAREPATTERNTYPE_PATTERN_TYPE.containing_type = _MALWAREPATTERNTYPE
DESCRIPTOR.message_types_by_name['MalwarePatternType'] = _MALWAREPATTERNTYPE

MalwarePatternType = _reflection.GeneratedProtocolMessageType('MalwarePatternType', (_message.Message,), dict(
  DESCRIPTOR = _MALWAREPATTERNTYPE,
  __module__ = 'MalwarePatternType_pb2'
  # @@protoc_insertion_point(class_scope:MalwarePatternType)
  ))
# _sym_db.RegisterMessage(MalwarePatternType)


# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python3

"""Keeps local Google Safe Browsing cache in sync.

Accessing Google Safe Browsing API requires API key, you can find
more info on getting it here:
https://developers.google.com/safe-browsing/lookup_guide#GettingStarted

"""
import argparse
import json  # FIXME: Needed?
import sys
import time
import os
import logging
log = logging.getLogger('gglsbl3')
try:
    from gglsbl3 import SafeBrowsingList
except ImportError:  # some magic to allow usage even when gglsbl3 is not installed (i.e. in the Python Path)
    try:  # trying relative import
        sys.path.append(os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(__file__)), "..")))
        # If the following fails, then something went wrong even with the relative import! A wrong folder structure may be a cause
        from gglsbl3 import SafeBrowsingList
    except ImportError:
        raise ImportError("Seems like gglsbl3 is not installed (or not in the right Folder or you are missing dependencies)")


def setupArgsParser():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('--api-key',
                        default=None,
                        required=True,
                        help='Safe Browsing API key [REQUIRED]')
    parser.add_argument('--db-path',
                        default='./gsb_v3.db',
                        help='Path to SQLite DB')
    parser.add_argument('--log',
                        default=None,
                        help='Path to log file, by default log to STDERR')
    parser.add_argument('--check-url',
                        default=None,
                        help='Check if URL is in black list and exit')
    parser.add_argument('--debug',
                        default=False,
                        action='store_true',
                        help='Show debug output')
    parser.add_argument('--onetime',
                        default=False,
                        action='store_true',
                        help='Run blacklists sync only once with reduced delays')
    # FXIME: add -h and --help | Edit: seems to be added by default, find a way to customise it
    # FIXME: create aliases
    return parser

# FIXME: move logging stuff to here. Beware of the scope
def setup_logger(log_file, debug):
    ch = logging.StreamHandler(sys.stdout)
    if debug:
        ch.setLevel(logging.DEBUG)
    else:
        ch.setLevel(logging.INFO)
    formatter = logging.Formatter('%(levelname)s: %(message)s - %(filename)s:%(lineno)d')
    ch.setFormatter(formatter)
    log.addHandler(ch)
    log.setLevel(logging.DEBUG)
    log.debug('DUMMY: setting up logging in file %s with debug = %s', log_file, debug)


def run_sync(sbl):
    """
    Synchronises the local database with the remote google servers.
    Takes a SafeBrowsingList Object as an argumant.
    """
    try:
        sbl.update_hash_prefix_cache()
    except (KeyboardInterrupt, SystemExit) as e:
        log.info('Shutting down')
        sys.exit(0)
    except Exception as e:
        log.exception('Failed to synchronize with GSB service: {}'.format(e))
        time.sleep(3)


def main():
    # FIXME: Exit more gracefully on urllib.error.URLError and other exceptions.
    # catch exceptions individually and provide info on how to fix them
    args_parser = setupArgsParser()
    args = args_parser.parse_args()
    setup_logger(args.log, args.debug)
    # FIXME: Sync before lookup?
    if args.check_url:
        # FIXME: check for validity of API KEY, e.g. min-length
        sbl = SafeBrowsingList(args.api_key, db_path=args.db_path)
        bl = sbl.lookup_url(args.check_url)
        if bl is None:
            print('%s is not blacklisted' % args.check_url)
        else:
            print('%s is blacklisted in %s' % (args.check_url, bl))
        sys.exit(0)
    if args.onetime:
        sbl = SafeBrowsingList(args.api_key, db_path=args.db_path, discard_fair_use_policy=True)
        run_sync(sbl)
    else:
        sbl = SafeBrowsingList(args.api_key, db_path=args.db_path)
        while True:
            run_sync(sbl)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
from .protocol import PrefixListProtocolClient, FullHashProtocolClient, URL
from .storage import SqliteStorage
import logging
import binascii

log = logging.getLogger('gglsbl3')


class SafeBrowsingList(object):
    """Interface for Google Safe Browsing API

    supporting partial update of the local cache.
    https://developers.google.com/safe-browsing/developers_guide_v3
    """

    def __init__(self, api_key, db_path='./gsb_v3.db', discard_fair_use_policy=False):
        # FIXME: missing docstring
        self.prefixListProtocolClient = PrefixListProtocolClient(api_key, discard_fair_use_policy=discard_fair_use_policy)
        self.fullHashProtocolClient = FullHashProtocolClient(api_key)
        self.storage = SqliteStorage(db_path)

    def _close_storage(self):
        """
        Close the connection to the database
        """
        self.storage.close()

    # FIXME: return True if there was something to update, return False if local database is in sync
    def update_hash_prefix_cache(self):
        "Sync locally stored hash prefixes with remote server"
        existing_chunks = self.storage.get_existing_chunks()
        log.debug("%d existing chunks in database", len(existing_chunks))
        response = self.prefixListProtocolClient.retrieve_missing_chunks(existing_chunks=existing_chunks)
        if response.reset_required:
            log.warning("Database reset is required!")
            self.storage.total_cleanup()
        try:
            self.storage.del_add_chunks(response.del_add_chunks)
            self.storage.del_sub_chunks(response.del_sub_chunks)
            log.debug("got %d chunks in response", len(response.chunks))
            for chunk in response.chunks:
                if self.storage.chunk_exists(chunk):
                    log.debug('chunk #%d of type %s exists in stored list %s, skipping',
                              chunk.chunk_number, chunk.chunk_type, chunk.list_name)
                    continue
                self.storage.store_chunk(chunk)
        except Exception as e:
            log.error("Encountered unknown error while updating hash prefix cache (%s)", e)
            log.warning("Rolling back database because of error")
            self.storage.db.rollback()
            raise

    def _sync_full_hashes(self, hash_prefix):
        "Sync full hashes starting with hash_prefix from remote server"
        if not self.storage.full_hash_sync_required(hash_prefix):
            log.debug('Cached full hash entries are still valid for "%s", no sync required.', binascii.hexlify(hash_prefix).decode("ascii"))
            return
        full_hashes = self.fullHashProtocolClient.get_hashes([hash_prefix])
        log.debug("got full hashes: %s", full_hashes)
        if not full_hashes:
            return
        self.storage.store_full_hashes(hash_prefix, full_hashes)

    def lookup_url(self, url):
        """
        Lookup a URL in the local database.
        Returns a list of "google list names" if the URL is found in one, otherwise None
        """
        looked_up = self.lookup_url_with_metadata(url)
        if looked_up:
            return [x["list"] for x in looked_up]
        else:
            return None

    def lookup_url_with_metadata(self, url):
        """
        Lookup a URL in the local database.
        Returns a list of dictionaries containing the list name and metadata if the URL is found in one,
        otherwise None
        """
        url_hashes = URL(url).hashes
        # log.debug(url_hashes)
        for url_hash in url_hashes:
            list_name = self._lookup_hash(url_hash)
            if list_name:
                return list_name
        return None

    def _lookup_hash(self, full_hash):
        """Lookup URL hash in blacklists

        Returns names of lists it was found in.
        """
        log.debug('looking up "%s"', binascii.hexlify(full_hash).decode("ascii"))
        hash_prefix = full_hash[0:4]
        log.debug('looking up hash prefix "%s"', binascii.hexlify(hash_prefix).decode("ascii"))
        try:
            if self.storage.lookup_hash_prefix(hash_prefix):
                self._sync_full_hashes(hash_prefix)
                return self.storage.lookup_full_hash(full_hash)
        except:
            self.storage.db.rollback()
            raise



'''
Created on May 7, 2015

@author: sk
'''

def prettify_seconds(seconds):
    """
    Prettifies seconds.
    Takes number of seconds (int) as input and returns a prettified string.

    Example:
    >>> prettify_seconds(342543)
    '3 days, 23 hours, 9 minutes and 3 seconds'
    """
    if seconds < 0:
        raise Exception("negative input not allowed")
    signs = {"s": {"singular": "second", "plural": "seconds", },
             "h": {"singular": "hour", "plural": "hours"},
             "min": {"singular": "minute", "plural": "minutes"},
             "d": {"singular": "day", "plural": "days"}
            }
    seperator = ", "
    last_seperator = " and "

    def get_sign(unit, value):
        if value == 1 or value == -1:
            return signs[unit]["singular"]
        else:
            return signs[unit]["plural"]

    days, remainder = divmod(seconds, 86400)
    hours, remainder = divmod(remainder, 3600)
    minutes, seconds = divmod(remainder, 60)
    daystext = "{} {}".format(days, get_sign("d", days)) if days else ""
    hourstext = "{} {}".format(hours, get_sign("h", hours)) if hours else ""
    minutestext = "{} {}".format(minutes, get_sign("min", minutes)) if minutes else ""
    if (not seconds) and (days or hours or minutes):
        secondstext = ""
    else:
        secondstext = "{} {}".format(seconds, get_sign("s", seconds))
    output_list = [daystext, hourstext, minutestext, secondstext]
    filtered = list(filter(None, output_list))
    if len(filtered) <= 2:
        output = last_seperator.join(filtered)
    else:
        output = seperator.join(filtered[:-1]) + last_seperator + filtered[-1]
    return output

def format_max_len(string_to_format, max_len=15, replace="[...]"):
    """
    Formats a string so len(format_max_length(string_to_format)) <= max_len
    If the string_to_format is longer than max_len, it replaces characters in the middle with [...]

    Example:
    >>> util.format_max_len('abcdefghijklmnopqrstuvwxyz')
    'abcde[...]vwxyz'
    """
    real_max_len = max_len-len(replace)  # needed to count the [...] in the ouput length
    if len(string_to_format) <= max_len:
        return string_to_format
    first = real_max_len // 2  # Insert the [...] in the (floored) middle
    last = real_max_len - first
    return string_to_format[0:first] + replace + string_to_format[len(string_to_format) - last:]



"""
Protocol module mainly for Python representations of objects from the google API
"""
import urllib.parse
import urllib.request
import urllib.error
import struct
import time
import random
import posixpath
import re
import hashlib
import socket
import binascii
import logging
from io import BytesIO

from gglsbl3 import util
from gglsbl3 import protobuf_pb2
from gglsbl3 import MalwarePatternType_pb2

log = logging.getLogger('gglsbl3')


class BaseProtocolClient(object):
    """
    Base class FullHashProtocolClient and PrefixListProtocolClient inherit from.
    """
    def __init__(self, api_key, discard_fair_use_policy=False):
        self.config = {
            "base_url": "https://safebrowsing.google.com/safebrowsing/",
            "lists": [
                "goog-malware-shavar",
                "googpub-phish-shavar",
                "goog-unwanted-shavar"
            ],
            "url_args": {
                "key": api_key,
                "appver": "0.1",
                "pver": "3.0",
                "client": "api"
            }
        }
        self.discard_fair_use_policy = discard_fair_use_policy
        self._next_call_timestamp = 0
        self._error_count = 0

    def set_next_call_timeout(self, delay):
        """
        Sets the timeout to be used for the next call
        """
        log.debug('Next query will be delayed %s seconds', delay)
        self._next_call_timestamp = int(time.time()) + delay

    def get_fair_use_delay(self):
        "Compute Server Query Delay according to Request Frequency policy"
        if self._error_count == 1:
            delay = 60
        elif self._error_count > 1:
            delay = 60 * min(480, random.randint(30, 60) * (2 ** (self._error_count - 2)))
        else:
            delay = self._next_call_timestamp - int(time.time())
        return delay

    def fair_use_delay(self):
        "Delay server query according to Request Frequency policy"
        delay = self.get_fair_use_delay()
        if delay < 0:
            log.error("got negative delay: '%s', will not sleep", delay)
        elif not self.discard_fair_use_policy:
            log.info('Sleeping for %s', util.prettify_seconds(delay))
            time.sleep(delay)
        else:
            log.debug("didn't sleep because of settings")

    def api_call(self, url, payload=None):
        "Perform a call to Safe Browsing API"
        log.debug("performing api call to %s with payload: %s", url, payload)
        if payload is None:
            payload = b''
        if isinstance(payload, str):
            payload = bytes(payload.encode("ascii"))
        request = urllib.request.Request(url, data=BytesIO(payload),
                                         headers={'Content-Length': len(payload)})
        try:
            response = urllib.request.urlopen(request)
        except urllib.error.HTTPError:
            self._error_count += 1
            raise
        self._error_count = 0
        return response.read()

    def make_url(self, service):
        "Generate Safe Browsing API URL"
        url = urllib.parse.urljoin(self.config['base_url'], service)
        query_params = '&'.join(['%s=%s' % (k, v) for k, v in list(self.config['url_args'].items())])
        url = '%s?%s' % (url, query_params)
        return url


class Chunk(object):

    "Represents content of Data-response chunk content"

    def __init__(self, decoded_chunk_data, list_name):
        self.list_name = list_name
        self.hashes = []
        self.chunk_number = None
        self.chunk_type = None
        self.prefix_length = None
        self._load_chunk(decoded_chunk_data)

    def _load_chunk(self, decoded_chunk_data):
        "Decode hash prefix entries"
        chunk_type = 'add'
        prefix_length = 4
        if decoded_chunk_data.chunk_type == 1:
            chunk_type = 'sub'
        if decoded_chunk_data.prefix_type == 1:
            prefix_length = 32
        hashes_str = decoded_chunk_data.hashes
        hashes_count = len(hashes_str) // prefix_length
        hashes = []
        for i in range(hashes_count):
            hashes.append(hashes_str[prefix_length * i:prefix_length * (i + 1)])
        self.hashes = hashes
        self.chunk_number = decoded_chunk_data.chunk_number
        self.chunk_type = chunk_type
        self.prefix_length = prefix_length


class DataResponse(object):

    """Contains information on what changes need to be made

    to the local copy of hash prefixes list
    """

    def __init__(self, raw_data):
        self.del_add_chunks = []
        self.del_sub_chunks = []
        self.reset_required = False
        self._parse_data(raw_data)

    def _parse_data(self, data):
        lists_data = {}
        current_list_name = None
        for l in data:
            l = l.strip()
            if not l:
                continue
            if l.startswith('i:'):
                current_list_name = l.strip()[2:]
                lists_data[current_list_name] = []
            elif l.startswith('u:'):
                url = l[2:]
                if not url.startswith('https://'):
                    url = 'https://%s' % url
                lists_data[current_list_name].append(url)
            elif l.startswith('r:'):
                log.warn("Reset is required!")
                self.reset_required = True
            elif l.startswith('ad:'):
                chunk_id = l.split(':')[1]
                self.del_add_chunks.append(chunk_id)
            elif l.startswith('sd:'):
                chunk_id = l.split(':')[1]
                self.del_sub_chunks.append(chunk_id)
            else:
                raise RuntimeError('Response line has unexpected prefix: "%s"', l)
        self.lists_data = lists_data

    def _unpack_chunks(self, chunkDataFH):
        "Unroll data chunk containing hash prefixes"
        # log.debug("unpacking chunk data: {data}".format(data=chunkDataFH.read()))
        decoded_chunks = []
        while True:
            packed_size = chunkDataFH.read(4)
            if len(packed_size) < 4:
                break
            size = struct.unpack(">L", packed_size)[0]
            chunk_data = chunkDataFH.read(size)
            decoded_chunk = protobuf_pb2.ChunkData()
            decoded_chunk.ParseFromString(chunk_data)
            decoded_chunks.append(decoded_chunk)
            # log.debug("sucessfully decoded chunk: %s", decoded_chunk)  # This produces way too much ouput
        log.debug("decoded %d chunks", len(decoded_chunks))
        return decoded_chunks

    def _fetch_chunks(self, url):
        "Download chunks of data containing hash prefixes"
        log.debug("fetching chunk %s", util.format_max_len(url, max_len=45))
        response = urllib.request.urlopen(url)
        return response

    @property
    def chunks(self):
        "Generator iterating through the server respones chunk by chunk"
        log.debug("accessing chunks")
        for list_name, chunk_urls in list(self.lists_data.items()):
            for chunk_url in chunk_urls:
                log.debug("processing chunk url: %s", util.format_max_len(chunk_url, max_len=45))
                packed_chunks = self._fetch_chunks(chunk_url)
                for chunk_data in self._unpack_chunks(packed_chunks):
                    # log.debug("chunk_data: {data}".format(data=chunk_data))
                    chunk = Chunk(chunk_data, list_name)
                    # log.debug("yielding {chunk}".format(chunk=chunk))
                    yield chunk


class PrefixListProtocolClient(BaseProtocolClient):

    def __init__(self, api_key, discard_fair_use_policy=False):
        super(PrefixListProtocolClient, self).__init__(api_key, discard_fair_use_policy)
        self.set_next_call_timeout(random.randint(0, 300))

    def get_lists(self):
        "Get available black/white lists"
        log.info('Fetching available lists')
        url = self.make_url('list')
        response = self.api_call(url)
        log.debug("got response %s", response)
        lists = [l.strip() for l in response.split()]
        return lists

    def _fetchData(self, existing_chunks):
        "Get references to data chunks containing hash prefixes"
        log.debug("chunks: %s", existing_chunks)
        self.fair_use_delay()
        url = self.make_url('downloads')
        payload = []
        for l in self.config['lists']:
            list_data = existing_chunks.get(l, {})
            if not list_data:
                payload.append('%s;' % l)
                continue
            list_data_cmp = []
            if 'add' in list_data:
                list_data_cmp.append('a:%s' % list_data['add'])
            if 'sub' in list_data:
                list_data_cmp.append('s:%s' % list_data['sub'])
            payload.append('%s;%s' % (l, ':'.join(list_data_cmp)))
        payload = '\n'.join(payload) + '\n'
        response = self.api_call(url, payload)
        return response

    def _preparse_data(self, data):
        log.debug('preparsing data (length %d)', len(data))
        data = data.decode("ascii")
        data = data.split('\n')
        next_delay = data.pop(0).strip()
        if not next_delay.startswith('n:'):
            raise RuntimeError('Expected poll interval as first line, got "%s"', next_delay)
        self.set_next_call_timeout(int(next_delay[2:]))
        return data

    def retrieve_missing_chunks(self, existing_chunks=None):
        """Get list of changes from the remote server

        and return them as DataResponse object
        """
        if existing_chunks is None:
            existing_chunks = {}
        log.info('Retrieving prefixes with %d existing chunks', len(existing_chunks))
        log.debug('existing_chunks: %s', existing_chunks)
        raw_data = self._fetchData(existing_chunks)
        log.info("raw data length: %d", len(raw_data))
        # log.debug("got raw data: %s", str(raw_data))  # this produces way too much output!
        preparsed_data = self._preparse_data(raw_data)
        d = DataResponse(preparsed_data)
        return d


class FullHashProtocolClient(BaseProtocolClient):

    def fair_use_delay(self):
        """Throttle queries according to Request Frequency policy

        https://developers.google.com/safe-browsing/developers_guide_v3#RequestFrequency
        """
        delay = self.get_fair_use_delay()
        log.debug("preparing to sleep for %d seconds", delay)
        if delay > 0 and not self.discard_fair_use_policy:
            log.info('Sleeping for %s seconds', delay)
            time.sleep(delay)
        else:
            log.debug("didn't sleep because of settings. fair use is: %s",
                      self.discard_fair_use_policy)

    def get_fair_use_delay(self):
        if self._error_count > 1:
            delay = min(120, 30 * (2 ** (self._error_count - 2)))
        else:
            delay = self._next_call_timestamp - int(time.time())
        log.debug("delay returned is %d", delay)
        return delay

    def _parse_hash_entry(self, hash_entry):
        "Parse full-sized hash entry"
        log.debug("parsing hash entry for %s", hash_entry)
        hashes = {}
        metadata = {}
        while True:
            if not hash_entry:
                break
            has_metadata = False
            header, hash_entry = hash_entry.split(b'\n', 1)
            opts = header.split(b':')
            if len(opts) == 4:
                if opts[3] == b'm':
                    has_metadata = True
                else:
                    raise RuntimeError('Failed to parse full hash entry header "%s"' % header)
            list_name = opts[0]
            entry_len = int(opts[1])
            entry_count = int(opts[2])
            hash_strings = []
            metadata_strings = []
            for i in range(entry_count):
                hash_string = hash_entry[entry_len * i:entry_len * (i + 1)]
                hash_strings.append(hash_string)
            hash_entry = hash_entry[entry_count * entry_len:]
            if has_metadata:
                for i in range(entry_count):
                    next_metadata_len, hash_entry = hash_entry.split(b'\n', 1)
                    next_metadata_len = int(next_metadata_len)
                    metadata_str = hash_entry[:next_metadata_len]
                    metadata_strings.append(metadata_str)
                    hash_entry = hash_entry[next_metadata_len:]
            elif hash_entry:
                raise RuntimeError('Hash length does not match header declaration (no metadata)')
            hashes[list_name] = hash_strings
            log.debug("metadata strings are: %s", metadata_strings)
            #  now decode the metadata_strings with protobuf
            metadata_strings_parsed = []
            if has_metadata and metadata_strings:
                for metadata_string in metadata_strings:
                    try:
                        metadata_string_proto = MalwarePatternType_pb2.MalwarePatternType()
                        metadata_string_proto.ParseFromString(metadata_string)
                        metadata_string_parsed = metadata_string_proto.pattern_type  # pylint: disable=E1101
                        metadata_strings_parsed.append(metadata_string_parsed)
                    except Exception:
                        log.error("failed to parse metadata string: '%s'", metadata_string)
                        raise
            else:
                log.warn("hash '%s' has not metadata!", hash_entry)
            log.debug("parsed metadata strings: %s", metadata_strings_parsed)
            metadata[list_name] = metadata_strings_parsed
        return hashes, metadata

    def get_hashes(self, hash_prefixes):
        "Download and parse full-sized hash entries"
        debug_prefixes = [binascii.hexlify(hash_prefix).decode("ascii") for hash_prefix in hash_prefixes]
        log.info('Downloading hashes for hash prefixes %s', debug_prefixes)
        url = self.make_url('gethash')
        prefix_len = len(hash_prefixes[0])
        hashes_len = prefix_len * len(hash_prefixes)
        p_header = '%d:%d' % (prefix_len, hashes_len)
        #  p_body = ''.join(hash_prefixes)
        p_body = b''
        for item in hash_prefixes:
            p_body += item
        payload = bytes(p_header.encode("ascii")) + b'\n' + p_body
        #  payload = '%s\n%s' % (p_header, p_body)

        response = self.api_call(url, payload)
        log.debug("response: %s", str(response))
        first_line, response = response.split(b'\n', 1)
        cache_lifetime = int(first_line.strip())
        hashes, metadata = self._parse_hash_entry(response)
        log.debug("got metadata: %s", metadata)
        return {'hashes': hashes,
                'metadata': metadata,
                'cache_lifetime': cache_lifetime
               }


class URL(object):

    "URL representation suitable for lookup"

    def __init__(self, url):
        self.url = str(url)

    @property
    def hashes(self):
        "Hashes of all possible permutations of the URL in canonical form"
        for url_variant in self.url_permutations(self.canonical):
            url_hash = self.digest(url_variant)
            yield url_hash

    @property
    def canonical(self):
        "Convert URL to its canonical form"
        def full_unescape(u):
            """
            Undo escaping of special characters in url
            """
            uu = urllib.parse.unquote(u)
            if uu == u:
                return uu
            else:
                return full_unescape(uu)

        def quote(unsafe_string):
            """
            Returns url safe representation of input with special characters escaped
            """
            safe_chars = '!"$&\'()*+,-./:;<=>?@[\\]^_`{|}~'
            return urllib.parse.quote(unsafe_string, safe=safe_chars)
        url = self.url.strip()
        url = url.replace('\n', '').replace('\r', '').replace('\t', '')
        url = url.split('#', 1)[0]
        url = quote(full_unescape(url))
        url_parts = urllib.parse.urlsplit(url)
        if not url_parts[0]:
            url = 'http://%s' % url
            url_parts = urllib.parse.urlsplit(url)
        protocol = url_parts.scheme
        host = full_unescape(url_parts.hostname)
        path = full_unescape(url_parts.path)
        query = url_parts.query
        if not query and '?' not in url:
            query = None
        if not path:
            path = '/'
        has_trailing_slash = (path[-1] == '/')
        path = posixpath.normpath(path).replace('//', '/')
        if has_trailing_slash and path[-1] != '/':
            path = path + '/'
        _user = url_parts.username
        port = url_parts.port
        host = host.strip('.')
        host = re.sub(r'\.+', '.', host).lower()
        if host.isdigit():
            log.debug("Host is digit: %s", host)
            try:
                host = util.int_to_ip(int(host))
                log.debug("after gethostbyname host is now %s", host)
            except socket.gaierror as e:
                log.warning("gethostbyname failed for host %s (%s)", host, e)
                pass
        if host.startswith('0x') and '.' not in host:
            try:
                host = socket.gethostbyname(host)
            except socket.gaierror as e:
                log.warning("gethostbyname failed for host %s (%s)", host, e)
                pass
        if path == '':
            path = '/'
        quoted_path = quote(path)
        quoted_host = quote(host)
        if port is not None:
            quoted_host = '%s:%s' % (quoted_host, port)
        canonical_url = '%s://%s%s' % (protocol, quoted_host, quoted_path)
        if query is not None:
            canonical_url = '%s?%s' % (canonical_url, query)
        return canonical_url

    # FIXME: move these to own module and out of class
    @staticmethod
    def url_permutations(url):
        """Try all permutations of hostname and path which can be applied
        to blacklisted URLs"""
        def url_host_permutations(host):
            """
            Generator to get all the allowed permutations for the host part of the url
            """
            if re.match(r'\d+\.\d+\.\d+\.\d+', host):
                yield host
                return
            parts = host.split('.')
            l = min(len(parts), 5)
            if l > 4:
                yield host
            for i in range(l - 1):
                yield '.'.join(parts[i - l:])

        def url_path_permutations(path):
            """
            Generator to get all the allowed permutations of the path part of the url
            """
            if path != '/':
                yield path
            query = None
            if '?' in path:
                path, query = path.split('?', 1)
            if query is not None:
                yield path
            path_parts = path.split('/')[0:-1]
            curr_path = ''
            for i in range(min(4, len(path_parts))):
                curr_path = curr_path + path_parts[i] + '/'
                yield curr_path
        _protocol, address_str = urllib.parse.splittype(url)
        host, path = urllib.parse.splithost(address_str)
        _user, host = urllib.parse.splituser(host)
        host, _port = urllib.parse.splitport(host)
        host = host.strip('/')
        for h in url_host_permutations(host):
            for p in url_path_permutations(path):
                yield '%s%s' % (h, p)

    @staticmethod
    def digest(url):
        "Hash the URL"
        return hashlib.sha256(url.encode("ascii")).digest()



#!/usr/bin/env python3

from setuptools import setup

import sys
import os

if not sys.version[0]  == "3":
    raise Exception("This Program is for Python **VERSION 3** only!")

__version__ = '0.1.2'
here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the relevant file
with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='gglsbl3',
      version=__version__,
      description="Client library for Google Safe Browsing API",
      classifiers=[
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: Implementation :: CPython',
          'Intended Audience :: Developers',
          'Topic :: Internet',
          'Topic :: Software Development :: Libraries :: Python Modules',
      ],
      long_description=long_description,
      keywords='gglsbl3 gglsbl safebrowsing google-safe-browing googlesafebrowsing',
      author='Stefan Kuntz',
      author_email='Stefan.github@gmail.com',
      url='https://github.com/Stefan-Code/gglsbl3',
      license='Apache2',
      packages=['gglsbl3', 'gglsbl3.util'],
      install_requires=['argparse', 'python3-protobuf', ],
      scripts=['scripts/gglsbl_client.py'],
      )
